--=o=--=o=--=o=--=o=--=o=--=o=--=o=--=o=--=o=--=o=--=o=--=o=--=o=--=o=--=o=--
--|                                                                       |--
--|     Standard OOP Stuff                      Ralf Nieuwenhuijsen       |--
--|                                                                       |--
--|                                                                       |--
--|        For information about this library please read 'oop.txt'       |--
--|             You are using this file on your own risk.                 |--
--|                     This file is *freeware*                           |--
--|                                                                       |--
--=o=--=o=--=o=--=o=--=o=--=o=--=o=--=o=--=o=--=o=--=o=--=o=--=o=--=o=--=o=--

    global constant OOP_VERSION = 1
    
    constant TRUE = 1, FALSE = 0

--=o=--=o=--=o=--=o=--=o=--=o=--=o=--=o=--=o=--=o=--=o=--=o=--=o=--=o=--=o=--
--[Error Handling of a class]                  

--    without trace
    
    global sequence error_message error_message = "Unspecified error with an class attribute!\n"
    integer trace_error
    trace_error = FALSE

    function sprint(object x)
    sequence s 
    
        if atom(x) then
            return sprintf("%g", x)    
        else        
            s = "{"
            for i = 1 to length(x) do
                s = s & sprint(x[i])  
                if i < length(x) then
                    s = s & ','
                end if
            end for
            s = s & "}"
            if length(s) > 5 * 80 then
                s = s[1..(5*80)-10] & "..."
            end if
            return s    
        end if
        
    end function

    constant ERROR_CLASSLENGTH 
    = "A class should at least be of the length 4\n"
    & "The first three attributes should be:\n"
    & "  - a clone method:    MClone \n"
    & "  - a kill procedure:  PKill  \n"
    & "  - a name value:      VName  \n"
    & "It is followed by a type check sequence as the 4th element\n"
    & "It contains the routine pointers of the types of all attributes\n"
    & "It should have a length one less than the class sequence itself\n"
    & "Although, this last element should be automatically generated!"
    
    constant ERROR_CLASSTYPE   
    = "A class is a sequence containing the data of its attributes\n"
    & "The variable you passed was not an sequence."
    
    constant ERROR_POINTER
    = "A class can contain methods, procedures and functions\n"
    & "But actually we store the routine pointers as attributes\n"
    & "The routine pointer you gave us can not be executed!\n"
    & "There is a great chance you have misspelled the routine name.\n"
    & "Or you might have tried to store a value not retured by routine_id ()"
    
    constant ERROR_NAME
    = "Every class has a name, which is stored in the 3rd element.\n"
    & "You can acces this attribute with the constant VName\n"
    & "However, to ensure this name is nicely viewable in both Win32 and Dos32\n"
    & " you may only use characters with ascII values between 32 and 126\n"
    & "Off course, this name attribute is a one-dimensional sequence."
    
    include machine.e       -- crash_message ()
    
    function error (sequence mes, sequence n, object x)
        mes = "[ERROR!!]\n"
            & "Use the procedure oop_trace (1) to turn on tracing when a class error occurs\n"
            & "The name of the class: " & n & "\n"
            & "The data causing this failure: " & sprint (x) & "\n\n"
            
            & "Cause of error:\n"
            & mes 
            & "\n\n"
            & "If you are certain this should be considered legal, please contact me:\n"
            & " [Ralf Nieuwenhuijsen] - [nieuwen@xs4all.nl] \n"
            & "Don't forget to include the ex.err file!\n\n"   
            
        return mes
    end function
    
    global procedure oop_trace (integer i)
        trace_error = i
    end procedure

--=o=--=o=--=o=--=o=--=o=--=o=--=o=--=o=--=o=--=o=--=o=--=o=--=o=--=o=--=o=--
--[Local tools]
    
    type text (object x)    -- Displayable on both WIN32 and DOS32
        if not sequence (x) then
            return FALSE
        end if
        for index = 1 to length(x) do
            if sequence (x[index]) then
                error_message = ERROR_NAME
                return FALSE
            end if
            if x[index] < 32 or x[index] > 126 then
                error_message = ERROR_NAME
                return FALSE
            end if
        end for                              
        
        return TRUE
    end type    

--=o=--=o=--=o=--=o=--=o=--=o=--=o=--=o=--=o=--=o=--=o=--=o=--=o=--=o=--=o=--
--[Classes]

    -- This attributes are required!
    global constant MClone = 1,     -- Method that clones the class
                    PKill  = 2,     -- Procedure called before de-allocation
                    VName  = 3      -- A writable name for the class

    global type class (object x)
        
        if not sequence (x) then
            -- "Wrong" method of OOP 
            if trace_error then        
                puts (1, error (ERROR_CLASSTYPE, x[VName],x) )
                trace (1)
                return TRUE -- The value is actually illegal
            else
                crash_message ( error (ERROR_CLASSTYPE,x[VName],x) )
                return FALSE
            end if
        elsif length(x) < 4 then                        
            if trace_error then        
                puts (1, error (ERROR_CLASSLENGTH,x[VName],x) )
                trace (1)
                return TRUE -- The value is actually illegal
            else
                crash_message ( error (ERROR_CLASSLENGTH,x[VName],x) )
                return FALSE                          
            end if
        end if           
        
        for index = 1 to length(x) - 1 do
            if not call_func (x[length(x)][index],{x[index]}) then
                if trace_error then        
                    puts (1, error(error_message,x[VName],x[index]) )
                    trace (1)
                    return TRUE -- The value is actually illegal
                else
                    crash_message ( error(error_message,x[VName],x[index]) )
                    return FALSE
                end if
            end if
        end for   
        
        return TRUE
    end type
    
--[Empty Attributes]

    function nothing_method (class x)
        return x
    end function

    procedure nothing_procedure (class x)
        return
        x = x       -- Turns off the warning about 'x' !
    end procedure                                     
    
    function nothing_function (class x)
        return TRUE                                         
        x = x       -- Turns off the warning about 'x' !
    end function
    
    global constant MNul = routine_id ("nothing_method"),
                    PNul = routine_id ("nothing_procedure"),
                    FNul = routine_id ("nothing_function") -- Returns TRUE
                    
--[Programmer Interface]

    -- While creating and modeling the class, we store it here:
    class active
    
    -- First you declare your new class
    global procedure declare_class (text name, class super)
        active = call_func(super[MClone],{super})
        active[VName] = name
    end procedure
    
    -- You can assign new values to the inherited attributes
    global procedure assign_class (integer i, object val)
        active[i] = val
    end procedure
    
    -- You can also add attributes
    global function attrib (object x, integer checkid)
        
        active = append(append(active[1..length(active)-1], x), append(active[length(active)], checkid))
        
        return length(active) - 1
    end function   
    
    -- Then you have to store the class
    global function store_class ()
    
        return active
    end function
    
--[Attribute Types]

    type routine_pointer (integer i)
        if i < 0 then
            error_message = ERROR_POINTER
        end if            
        return i >= 0
    end type                        
    
    function always_right (object x)
        return TRUE                                   
        x = x
    end function
    
    -- The 4 basic types of attributes
    global constant Method = routine_id ("routine_pointer"),
                    Procedure = Method,
                    Function = Method,
                    Value = routine_id ("always_right"),
                    Class = routine_id ("class")
             
    -- Custom types: (are all basic types: Value)
    constant  Text = routine_id ("text")

--[Empty Class]

    global constant NewClass = 
        {
            MNul,
            PNul,
            "EmptyClass",
            
            { Method, Procedure, Text }
        }                              
        
--[Verification & Comparisements]

    global function compare_class (class a, class b)
    
        if length(a) < length(b) then
            if b[length(b)][1..length(a)-1] = a[length(a)] then
                return -1
            else
                return FALSE
            end if
        elsif length(a) > length(b) then
            if a[length(a)][1..length(b)-1] = b[length(b)] then
                return 1
            else        
                return FALSE
            end if
        else
            return FALSE
        end if
        
    end function
    
    global function equal_class (class a, class b)
    
        return a[length(a)] = b[length(b)]

    end function
    
    global function restore_class (class x, class super)
    
        if compare_class (x, super) != -1 
        and (compare_class (x,super) or equal_class (x, super)) then
            x[1..length(super)-1] = super[1..length(super)-1]
        end if
        
        return x
    end function                                   
    
    global function try_attrib (class x, integer at, object val)
    
        return call_func(x[length(x)][at], {val})
        
    end function              
    
    global procedure kill_class (class x)
        
        call_proc (x[PKill],{x})
        
    end procedure
    
    global function method_class (class x, integer i, sequence a)
        
        return call_func (x[i], {x} & a)
        
    end function                    
    
    global function value_class (class x, integer i, object z)
    
        x[i] = z        -- You could have done this yourself, dummy!@

        return x
    
    end function
    
    global procedure procedure_class (class x, integer i, sequence a)
    
        call_proc (x[i], {x} & a)
        
    end procedure
    
    global function function_class (class x, integer i, sequence a)
        
        return call_func (x[i], {x} & a)
        
    end function                  

